// maxlispj -- java interface to abcl lisp (http://armedbear.org/abcl.html)
//			(now: http://common-lisp.net/project/armedbear/)
// for max/msp.  This uses abcl v. 0.0.10.0 
//
// compile this with:
//		javac -classpath ".:[pathto]/abcl.jar:[pathto]/max.jar"
//
//		Brad Garton, 12/2007
//		v. 0.9
//
// 1/2011, v. 0.96
//	-- updated to abcl v. 0.0.23.0
// -- included bug fix for large LISP buffers on the object
// -- error reporting now does backtrace dump into Max window
// -- OSX 10.6, Java v. 1.6

import org.armedbear.lisp.*;
import java.io.*;
import java.lang.*;
import java.awt.Frame;
import java.awt.TextArea;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import com.cycling74.max.*;

public class maxlispj extends MaxObject implements WindowListener {
	Interpreter interpreter;
	PipedOutputStream pipeout;
	PipedInputStream pipein;
	Stream stdout;
	String lispbuf, evalstring;
	Frame bufframe;
	TextArea buftext;
	String lispfname;
	boolean noparens = false;
	boolean noquotes = false;

	public maxlispj() {
		declareOutlets(new int[] { DataTypes.MESSAGE, DataTypes.MESSAGE, DataTypes.MESSAGE });
		declareInlets(new int[] { DataTypes.MESSAGE });

		interpreter = Interpreter.createInstance();
		if (interpreter == null) interpreter = Interpreter.getInstance();

		// this is how we capture the stdout printing from abcl
		try {
			pipeout = new PipedOutputStream();
			pipein = new PipedInputStream();
			pipein.connect(pipeout);
			stdout = new Stream(Symbol.SYSTEM_STREAM, pipeout, Symbol.CHARACTER, true);
		} catch (Throwable t) {
				error("could not construct the lisp stdout pipe");
		}

		lispbuf = new String();

		post("maxlispj by Brad Garton, v 0.96, abcl v. 0.0.23.0");
	}

	public void bang() {
		lisp(lispbuf);
	}

	public void version() {
			outlet(0, "maxlispj by Brad Garton, v 0.96, abcl v. 0.0.23.0");
	}

	public void lisp(Atom[] a) {
		lisp(Atom.toOneString(a));
	}

	public void text(Atom[] a) {
		lisp(Atom.toOneString(a));
	}

	public void text(String s) {
		lisp(s);
	}

	public void lisp(String lbuf) {
		if (lbuf.length() == 0) {
			error("you are evaluating a 0-length lisp buffer");
			outlet(2, "ZERO_LENGTH_BUFFER");
			return;
		}

		Symbol.STANDARD_OUTPUT.setSymbolValue(stdout); // output to this instance
		int nparens = 0;
		int nlines = 0; // for attempt to report parens error line
		try {
			// go through the buffer and eval each lisp form
			for (int i = 0; i < lbuf.length(); i++) {

				if (lbuf.charAt(i) == ';') { // ignore comments	
					while (lbuf.charAt(++i) != '\n') {
						if (i == (lbuf.length() - 1)) break;
					}
					nlines++;
				}

				if (lbuf.charAt(i) == '\n') nlines++;

				if (lbuf.charAt(i) == ')') {
					error("unbalanced right parens near line " + nlines);
					outlet(2, "UNBALANCED_PARENS");
				}
				else if (lbuf.charAt(i) == '(') { // beginning of the form
					nparens = 1;
					int startdex = i;
					while (++i < lbuf.length()) {
						if (lbuf.charAt(i) == '(') nparens++;
						if (lbuf.charAt(i) == ')') nparens--;
						if (lbuf.charAt(i) == '\n') nlines++;
						if (nparens == 0) break;
					}

					if (nparens == 0) { // we have a form to evaluate
						int enddex = i; // the end of this form in the string buffer

						// to catch errors; the thread changes when invoked
						// thanks to Alex Mizrahi for the code!
						LispThread thread = LispThread.currentThread();
						Symbol dbg = Lisp.PACKAGE_SYS.findAccessibleSymbol("%DEBUGGER-HOOK-FUNCTION");
						LispObject throwingDebugHook = dbg.getSymbolFunction();
						thread.setSpecialVariable(Symbol.DEBUGGER_HOOK, throwingDebugHook);

						evalstring = interpreter.eval(lbuf.substring(startdex, enddex+1)).writeToString();
					} else {
						error("unbalanced left parens, ending at line " + nlines);
						outlet(2, "UNBALANCED_PARENS");
					}
				}
			}

			interpreter.eval("(terpri)"); // need this to flush the print buffer

			if (noparens) { // get rid of parens in eval output
				evalstring = evalstring.replace('(', ' ');
				evalstring = evalstring.replace(')', ' ');
				evalstring.trim();
			}

			if (noquotes) { // get rid of quotes in eval output
				evalstring = evalstring.replace('"', ' ');
				evalstring.trim();
			}
			outlet(0, evalstring);

			if (pipein.available() > 0) {
				byte [] bbb = new byte[pipein.available()];
				pipein.read(bbb);
				String sss = new String(bbb);
				if (noparens) { // get rid of parens in print output
					sss = sss.replace('(', ' ');
					sss = sss.replace(')', ' ');
					sss = sss.trim();
				}
				if (noquotes) {  // get rid of quotes in print output
					sss = sss.replace('"', ' ');
					sss = sss.trim();
				}
				outlet(1, sss);
			}
		} catch (Throwable c) { // there was a lisp eval error
			try {
				outlet(2, "LISP EVAL ERROR (see Max window for backtrace)");
				c.printStackTrace();
//				c.getMessage();
//				c.toString();
            System.out.println("LISP eval error, backtrace above");
			} catch (Throwable c2) {
				error("YIKES! error writing the error!");
				outlet(2, c2.getMessage());
			}
		}
	}

	public void noparens(String s) {
		if (s.startsWith("t")) noparens = true;
		else if (s.startsWith("f")) noparens = false;
	}

	public void noquotes(String s) {
		if (s.startsWith("t")) noquotes = true;
		else if (s.startsWith("f")) noquotes = false;
	}


// buffer/window methods
	public void openbuf() {
		bufframe = new Frame("lisp buffer");
		bufframe.setSize(500, 400);

		buftext = new TextArea(lispbuf);

		bufframe.add(buftext);
		bufframe.setVisible(true);
		bufframe.addWindowListener(this);
	}

	public void loadbuf(String lf) {
		if (lf.length() == 0)
			lispfname = MaxSystem.openDialog();
		else lispfname = lf;

		if (lispfname != null) {
			try {
				File ff = new File(lispfname);
				FileReader fr = new FileReader(ff);
				char [] cbuf = new char[(int)ff.length()];
				fr.read(cbuf);
				lispbuf = new String(cbuf);
			} catch (IOException o) {
				error("could not load file: " + lispfname);
			}
		}
	}

	public void savebuf(String lf) {
		if (lf.length() == 0)
			lispfname = MaxSystem.saveAsDialog("maxlispj buffer", "*.lisp");
		else lispfname = lf;

		if (lispfname != null) {
			try {
				File ff = new File(lispfname);
				FileWriter fw = new FileWriter(ff);
				fw.write(lispbuf);
				fw.flush();
				fw.close();
			} catch (IOException o) {
				error("could not save file: " + lispfname);
			}
		}
	}

	public void save() { // save the lisp buffer for this object
		int lowdex, hidex;

		// split it up into small chunks so that buffers > 2k may be saved
		// the first atom has the number of chunks (atoms w/ strings)

// BGG mmx -- arg, this is so annoying.  Max/MSP escapes quote chars with
//		a backslash in the .maxpat XML file, so this means that repeated
//		saves increase the number of backslashes each time.  Oh what fun.
//		So I replace the backslash with a marker "xxDQxx" and then do the
//		inverse when restoring
		lispbuf = lispbuf.replace("\"", "xxDQxx");
		lispbuf = lispbuf.replace(";", "xxSCxx");
		lispbuf = lispbuf.replace(",", "xxCMxx");
		lispbuf = lispbuf.replace("\\", "xxBSxx");

		if (lispbuf.length() > 0) {
// BGG mmx -- I had to increase the segmenting because there appears
//		to be an upper bound on # of atoms in embedMessage().  I think
//		maybe I may not need to segment at all (1000000 works ok!)
			int natoms = ((int) (lispbuf.length()/10000)) + 1;
			Atom [] saveatoms = new Atom[natoms+1];
			saveatoms[0] = Atom.newAtom(natoms);
			for (int i = 0; i < natoms; i++) {
				lowdex = i * 10000;
				hidex = (lowdex + 10000);
				if (hidex > lispbuf.length()) hidex = lispbuf.length();
				saveatoms[i+1] =  Atom.newAtom(lispbuf.substring(lowdex, hidex));
			}
			embedMessage("setbuf", saveatoms);

		}
	}

	public void setbuf(Atom [] a) {
		// see the comment in the save() method above
		int natoms = a[0].getInt();
		for (int i = 0; i < natoms; i++) {
			lispbuf = lispbuf.concat(a[i+1].getString());
		}

// BGG mmx -- see note above about backslashes in save()
	lispbuf = lispbuf.replace("xxDQxx", "\"");
	lispbuf = lispbuf.replace("xxSCxx", ";");
	lispbuf = lispbuf.replace("xxCMxx", ",");
	lispbuf = lispbuf.replace("xxBSxx", "\\");

	}


	public void windowClosing(WindowEvent e) {
		lispbuf = buftext.getText();
		bufframe.dispose();
	}
	public void windowActivated(WindowEvent e) { }
	public void windowClosed(WindowEvent e) { }
	public void windowDeactivated(WindowEvent e) { }
	public void windowDeiconified(WindowEvent e) { }
	public void windowIconified(WindowEvent e) { }
	public void windowOpened(WindowEvent e) { }
}
